@@ -1,3 +1,4 @@
+import pdb
 import tblib.pickling_support
 # allows pickling of basic information from tracebacks:
 tblib.pickling_support.install()
@@ -84,14 +85,6 @@ class EnumeratedItem(object):
         return iter((self.number, self.value))
 
 
-def unpack(enumerated_args):
-    new_enumeration = tuple(i for (i, __) in enumerated_args)
-    if len(new_enumeration) == 1:
-        new_enumeration = new_enumeration[0]
-    argument = tuple(a for (__, a) in enumerated_args)
-    return new_enumeration, argument
-
-
 def wrap_callback(new_enumeration, callback):
     """creates a new callback which builds a EnumeratedItem from the reult """
     def new_callback(result):
@@ -128,9 +121,9 @@ class Engine(object):
         assert all(isinstance(arg, EnumeratedItem) for arg in enumerated_args)
         for arg in enumerated_args:
             if isinstance(arg.value, DelayedException):
-                callback(arg)
+                callback(arg.value)
                 return
-        new_enumeration, args = unpack(enumerated_args)
+        args = tuple(a for (__, a) in enumerated_args)
         # dill can pickle more function types as pickle module (which underlies
         # multiprocessing) , eg lambdas or static methods, so we transport our computation job
         # serialized as strings, which we deserialize in the remote process again:
@@ -140,10 +133,10 @@ class Engine(object):
         # for local and remote execution. this supports debugging.
         if Engine.pool is None:
             result = eval_pickled(f_s, args_s)
-            wrap_callback(new_enumeration, callback)(result)
+            callback(result)
         else:
             Engine.pool.apply_async(eval_pickled, (f_s, args_s),
-                                    callback=wrap_callback(new_enumeration, callback))
+                                    callback=callback)
 
 
 class DataStream(object):
@@ -243,18 +236,20 @@ class _UpdateManagerForJoin(object):
 
     def __init__(self, input_streams):
         self.seen = set()
-        # a list of collected data per stream:
-        self.data = [list() * a.size() for a in input_streams]
+        self.data = [[None for __ in range(a.size())] for a in input_streams]
+        self.next_ = [0] * len(input_streams)
 
     def new_function_arguments_for(self, stream_index, item):
-        self.data[stream_index].append(item)
+        self.data[stream_index][self.next_[stream_index]] = item
+        self.next_[stream_index] += 1
         # cross product over all indices of available data:
         sizes = [len(v) for v in self.data]
-        for perm in itertools.product(*(range(size) for size in sizes)):
+        for number, perm in enumerate(itertools.product(*(range(size) for size in sizes))):
             if perm not in self.seen:
                 args = tuple(self.data[j][pi] for j, pi in enumerate(perm))
-                yield args
-                self.seen.add(perm)
+                if all(ai is not None for ai in args):
+                    yield number, args
+                    self.seen.add(perm)
 
     def input_data_pending(self, size):
         return len(self.seen) < size
@@ -276,8 +271,8 @@ class JoinComputation(Computation):
                     if isinstance(item.value, DelayedException):
                         self.put(item)
                         return
-                    for args in manager.new_function_arguments_for(i, item):
-                        Engine.run_async(self.f, args, callback=self.put)
+                    for number, args in manager.new_function_arguments_for(i, item):
+                        Engine.run_async(self.f, args, callback=wrap_callback(number, self.put))
                         started_computation = True
             if not started_computation:
                 time.sleep(0.001)
@@ -304,7 +299,7 @@ class SummarizeComputation(Computation):
         items.sort(key=lambda item: item.number)
         values = [v for __, v in items]
         arg = EnumeratedItem(0, values)
-        Engine.run_async(self.f, (arg,), callback=self.put)
+        Engine.run_async(self.f, (arg,), callback=wrap_callback(0,self.put))
 
 
 class ZipComputation(Computation):
@@ -339,17 +334,17 @@ class ZipComputation(Computation):
                     return True
         return False
 
-    def _assemble_function_arguments(self, collected_input_data, cip):
+    def _assemble_function_arguments(self, collected_inputs, seen):
         for i, stream in enumerate(self.input_streams):
             if not stream.empty():
                 item = stream.get()
-                if item.number not in collected_input_data.keys():
-                    collected_input_data[item.number] = cip
-                collected_input_data[item.number][i] = item
-        for key, arg in collected_input_data.items():
-            if all(c is not None for c in arg):
-                yield arg
-            del collected_input_data[key]
+                collected_inputs[item.number][i] = item
+
+        for ni, arg in enumerate(collected_inputs):
+            if ni not in seen:
+                if all(c is not None for c in arg):
+                    yield ni, arg
+                    seen.add(ni)
 
     def _start_computation(self):
         n = 0
@@ -360,20 +355,26 @@ class ZipComputation(Computation):
                 if isinstance(item.value, DelayedException):
                     self.put(item)
                     return
-        collected_input_data = dict()
         # only constant inputs ?
         if self.size() == 1:
-            Engine.run_async(self.f, cip, callback=self.put)
+            Engine.run_async(self.f, cip, callback=wrap_callback(0, self.put))
         else:
+            collected_inputs = [[None for __ in self.input_streams] for k in range(self.size())]
+            for i, is_ in enumerate(self.input_streams):
+                if is_.size() == 1:
+                    for ii in range(self.size()):
+                        collected_inputs[ii][i] = cip[i]
+
+            seen = set()
             while n < self.size():
                 avail = self._check_for_any_new_data()
                 if avail:
-                    for args in self._assemble_function_arguments(collected_input_data, cip):
+                    for number, args in self._assemble_function_arguments(collected_inputs, seen):
                         for arg in args:
                             if isinstance(arg, DelayedException):
-                                self.put(arg)
+                                wrap_callback(number, self.put)(arg)
                                 return
-                        Engine.run_async(self.f, args, callback=self.put)
+                        Engine.run_async(self.f, args, callback=wrap_callback(number, self.put))
                         n += 1
                 else:
                     time.sleep(0.001)
